@@ -2,11 +2,14 @@ use std::{convert::Infallible, str::Utf8Error, string::FromUtf8Error, sync::Arc}
 
 use bonsaidb_core::{
     permissions::PermissionDenied,
-    schema::{view, InvalidNameError},
+    pubsub::{Disconnected, TryReceiveError},
+    schema::{view, InsertError, InvalidNameError},
     AnyError,
 };
 use nebari::AbortError;
 
+use crate::database::compat::UnknownVersion;
+
 /// Errors that can occur from interacting with storage.
 #[derive(thiserror::Error, Debug)]
 pub enum Error {
@@ -16,7 +19,7 @@ pub enum Error {
 
     /// An error occurred serializing the underlying database structures.
     #[error("error while serializing internal structures: {0}")]
-    InternalSerialization(#[from] bincode::Error),
+    InternalSerialization(String),
 
     /// An error occurred serializing the contents of a `Document` or results of a `View`.
     #[error("error while serializing: {0}")]
@@ -26,6 +29,10 @@ pub enum Error {
     #[error("error while communicating internally")]
     InternalCommunication,
 
+    /// A transaction was too large to execute.
+    #[error("transaction is too large")]
+    TransactionTooLarge,
+
     /// An error occurred while executing a view
     #[error("error from view: {0}")]
     View(#[from] view::Error),
@@ -50,12 +57,13 @@ pub enum Error {
     Core(#[from] bonsaidb_core::Error),
 
     /// A tokio task failed to execute.
+    #[cfg(feature = "async")]
     #[error("a concurrency error ocurred: {0}")]
     TaskJoin(#[from] tokio::task::JoinError),
 
-    /// A tokio task failed to execute.
+    /// An io error occurred.
     #[error("an IO error occurred: {0}")]
-    Io(#[from] tokio::io::Error),
+    Io(#[from] std::io::Error),
 
     /// An error occurred from a job and couldn't be unwrapped due to clones.
     #[error("an error from a job occurred: {0}")]
@@ -71,12 +79,42 @@ pub enum Error {
     PasswordHash(String),
 }
 
+impl<T> From<InsertError<T>> for Error {
+    fn from(err: InsertError<T>) -> Self {
+        Self::Core(err.error)
+    }
+}
+
 impl From<flume::RecvError> for Error {
     fn from(_: flume::RecvError) -> Self {
         Self::InternalCommunication
     }
 }
 
+impl From<TryReceiveError> for Error {
+    fn from(_: TryReceiveError) -> Self {
+        Self::InternalCommunication
+    }
+}
+
+impl From<Disconnected> for Error {
+    fn from(_: Disconnected) -> Self {
+        Self::InternalCommunication
+    }
+}
+
+impl From<bincode::Error> for Error {
+    fn from(err: bincode::Error) -> Self {
+        Self::InternalSerialization(err.to_string())
+    }
+}
+
+impl<T> From<UnknownVersion<T>> for Error {
+    fn from(err: UnknownVersion<T>) -> Self {
+        Self::InternalSerialization(err.to_string())
+    }
+}
+
 #[cfg(feature = "password-hashing")]
 impl From<argon2::Error> for Error {
     fn from(err: argon2::Error) -> Self {
@@ -91,12 +129,14 @@ impl From<argon2::password_hash::Error> for Error {
     }
 }
 
+#[cfg(feature = "async")]
 impl From<tokio::sync::oneshot::error::RecvError> for Error {
     fn from(_: tokio::sync::oneshot::error::RecvError) -> Self {
         Self::InternalCommunication
     }
 }
 
+#[cfg(feature = "async")]
 impl From<tokio::sync::oneshot::error::TryRecvError> for Error {
     fn from(_: tokio::sync::oneshot::error::TryRecvError) -> Self {
         Self::InternalCommunication
