@@ -22,7 +22,6 @@ pub mod cli;
 pub mod config;
 mod database;
 mod error;
-pub mod jobs;
 mod open_trees;
 mod storage;
 mod tasks;
@@ -36,10 +35,16 @@ pub use argon2;
 pub use bonsaidb_core as core;
 
 pub use self::{
-    database::{pubsub::Subscriber, Database},
+    database::{pubsub::Subscriber, Database, DatabaseNonBlocking},
     error::Error,
-    storage::{BackupLocation, Storage, StorageId},
+    storage::{BackupLocation, Storage, StorageId, StorageNonBlocking},
 };
 
+#[cfg(feature = "async")]
+mod r#async;
+
+#[cfg(feature = "async")]
+pub use r#async::*;
+
 #[cfg(test)]
 mod tests;
