@@ -1,30 +1,30 @@
 use std::{
+    fs::DirEntry,
     io::ErrorKind,
     path::{Path, PathBuf},
 };
 
-use async_trait::async_trait;
 use bonsaidb_core::{
     admin,
-    connection::{Connection, Range, Sort, StorageConnection},
+    connection::{LowLevelConnection, Range, Sort, StorageConnection},
     document::DocumentId,
-    schema::{Collection, SchemaName},
+    schema::{Collection, Qualified, SchemaName},
     transaction::{Operation, Transaction},
     AnyError,
 };
-use futures::Future;
-use tokio::fs::DirEntry;
 
-use crate::{database::keyvalue::Entry, Database, Error, Storage};
+use crate::{
+    database::{keyvalue::Entry, DatabaseNonBlocking},
+    Database, Error, Storage,
+};
 
 /// A location to store and restore a database from.
-#[async_trait]
 pub trait BackupLocation: Send + Sync {
     /// The error type for the backup location.
     type Error: AnyError;
 
     /// Store `object` at `path` with `name`.
-    async fn store(
+    fn store(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -34,13 +34,13 @@ pub trait BackupLocation: Send + Sync {
     ) -> Result<(), Self::Error>;
 
     /// Lists all of the schemas stored in this backup location.
-    async fn list_schemas(&self) -> Result<Vec<SchemaName>, Self::Error>;
+    fn list_schemas(&self) -> Result<Vec<SchemaName>, Self::Error>;
 
     /// List all of the names of the databases stored for `schema`.
-    async fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Self::Error>;
+    fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Self::Error>;
 
     /// List all stored named objects at `path`. The names should be the same that were provided when `store()` was called.
-    async fn list_stored(
+    fn list_stored(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -48,7 +48,7 @@ pub trait BackupLocation: Send + Sync {
     ) -> Result<Vec<String>, Self::Error>;
 
     /// Load a previously stored object from `path` with `name`.
-    async fn load(
+    fn load(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -59,90 +59,88 @@ pub trait BackupLocation: Send + Sync {
 
 impl Storage {
     /// Stores a copy of all data in this instance to `location`.
-    pub async fn backup<L: AnyBackupLocation>(&self, location: L) -> Result<(), Error> {
+    pub fn backup<L: AnyBackupLocation>(&self, location: &L) -> Result<(), Error> {
         let databases = {
-            self.data
+            self.instance
+                .data
                 .available_databases
                 .read()
-                .await
                 .keys()
                 .cloned()
                 .collect::<Vec<_>>()
         };
 
         for name in databases {
-            let database = self.database_without_schema(&name).await?;
-            self.backup_database(&database, &location).await?;
+            let database = self
+                .instance
+                .database_without_schema(&name, Some(self), None)?;
+            Self::backup_database(&database, location)?;
         }
 
         Ok(())
     }
 
-    /// Stores a copy of all data in this instance to `location`.
-    pub async fn restore<L: AnyBackupLocation>(&self, location: L) -> Result<(), Error> {
+    /// Restores all data from a previously stored backup `location`.
+    pub fn restore<L: AnyBackupLocation>(&self, location: &L) -> Result<(), Error> {
         for schema in location
             .list_schemas()
-            .await
             .map_err(|err| Error::Backup(Box::new(err)))?
         {
             for database in location
                 .list_databases(&schema)
-                .await
                 .map_err(|err| Error::Backup(Box::new(err)))?
             {
                 // The admin database is already going to be created by the process of creating a database.
-                self.create_database_with_schema(&database, schema.clone(), true)
-                    .await?;
+                self.create_database_with_schema(&database, schema.clone(), true)?;
 
-                let database = self.database_without_schema(&database).await?;
-                self.restore_database(&database, &location).await?;
+                let database =
+                    self.instance
+                        .database_without_schema(&database, Some(self), None)?;
+                Self::restore_database(&database, location)?;
             }
         }
 
         Ok(())
     }
 
-    pub(crate) async fn backup_database(
-        &self,
+    pub(crate) fn backup_database(
         database: &Database,
         location: &dyn AnyBackupLocation,
     ) -> Result<(), Error> {
         let schema = database.schematic().name.clone();
         for collection in database.schematic().collections() {
-            let documents = database
-                .list(Range::from(..), Sort::Ascending, None, &collection)
-                .await?;
+            let documents = database.list_from_collection(
+                Range::from(..),
+                Sort::Ascending,
+                None,
+                &collection,
+            )?;
             let collection_name = collection.encoded();
             // TODO consider how to best parallelize -- perhaps a location can opt into parallelization?
             for document in documents {
-                location
-                    .store(
-                        &schema,
-                        database.name(),
-                        &collection_name,
-                        &document.header.id.to_string(),
-                        &document.contents,
-                    )
-                    .await?;
+                location.store(
+                    &schema,
+                    database.name(),
+                    &collection_name,
+                    &document.header.id.to_string(),
+                    &document.contents,
+                )?;
             }
-            for ((namespace, key), entry) in database.all_key_value_entries().await? {
+            for ((namespace, key), entry) in database.all_key_value_entries()? {
                 let full_name = format!("{}._key._{}", namespace.as_deref().unwrap_or(""), key);
-                location
-                    .store(
-                        &schema,
-                        database.name(),
-                        "_kv",
-                        &full_name,
-                        &pot::to_vec(&entry)?,
-                    )
-                    .await?;
+                location.store(
+                    &schema,
+                    database.name(),
+                    "_kv",
+                    &full_name,
+                    &pot::to_vec(&entry)?,
+                )?;
             }
         }
         Ok(())
     }
 
-    pub(crate) async fn restore_database(
-        &self,
+    pub(crate) fn restore_database(
         database: &Database,
         location: &dyn AnyBackupLocation,
     ) -> Result<(), Error> {
@@ -160,8 +158,7 @@ impl Storage {
         {
             let collection_name = collection.encoded();
             for (id, id_string) in location
-                .list_stored(&schema, database.name(), &collection_name)
-                .await?
+                .list_stored(&schema, database.name(), &collection_name)?
                 .into_iter()
                 .filter_map(|id_string| {
                     id_string
@@ -170,29 +167,23 @@ impl Storage {
                         .map(|id| (id, id_string))
                 })
             {
-                let contents = location
-                    .load(&schema, database.name(), &collection_name, &id_string)
-                    .await?;
+                let contents =
+                    location.load(&schema, database.name(), &collection_name, &id_string)?;
                 transaction.push(Operation::insert(collection.clone(), Some(id), contents));
             }
         }
-        database.apply_transaction(transaction).await?;
+        database.apply_transaction(transaction)?;
 
-        for full_key in location
-            .list_stored(&schema, database.name(), "_kv")
-            .await?
-        {
+        for full_key in location.list_stored(&schema, database.name(), "_kv")? {
             if let Some((namespace, key)) = full_key.split_once("._key._") {
-                let entry = location
-                    .load(&schema, database.name(), "_kv", &full_key)
-                    .await?;
+                let entry = location.load(&schema, database.name(), "_kv", &full_key)?;
                 let entry = pot::from_slice::<Entry>(&entry)?;
                 let namespace = if namespace.is_empty() {
                     None
                 } else {
                     Some(namespace.to_string())
                 };
-                entry.restore(namespace, key.to_string(), database).await?;
+                entry.restore(namespace, key.to_string(), database)?;
             }
         }
 
@@ -200,9 +191,8 @@ impl Storage {
     }
 }
 
-#[async_trait]
 pub trait AnyBackupLocation: Send + Sync {
-    async fn store(
+    fn store(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -211,18 +201,18 @@ pub trait AnyBackupLocation: Send + Sync {
         object: &[u8],
     ) -> Result<(), Error>;
 
-    async fn list_schemas(&self) -> Result<Vec<SchemaName>, Error>;
+    fn list_schemas(&self) -> Result<Vec<SchemaName>, Error>;
 
-    async fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Error>;
+    fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Error>;
 
-    async fn list_stored(
+    fn list_stored(
         &self,
         schema: &SchemaName,
         database_name: &str,
         container: &str,
     ) -> Result<Vec<String>, Error>;
 
-    async fn load(
+    fn load(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -231,13 +221,12 @@ pub trait AnyBackupLocation: Send + Sync {
     ) -> Result<Vec<u8>, Error>;
 }
 
-#[async_trait]
 impl<L, E> AnyBackupLocation for L
 where
     L: BackupLocation<Error = E>,
     E: AnyError,
 {
-    async fn store(
+    fn store(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -246,34 +235,30 @@ where
         object: &[u8],
     ) -> Result<(), Error> {
         self.store(schema, database_name, container, name, object)
-            .await
             .map_err(|err| Error::Backup(Box::new(err)))
     }
 
-    async fn list_schemas(&self) -> Result<Vec<SchemaName>, Error> {
+    fn list_schemas(&self) -> Result<Vec<SchemaName>, Error> {
         self.list_schemas()
-            .await
             .map_err(|err| Error::Backup(Box::new(err)))
     }
 
-    async fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Error> {
+    fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Error> {
         self.list_databases(schema)
-            .await
             .map_err(|err| Error::Backup(Box::new(err)))
     }
 
-    async fn list_stored(
+    fn list_stored(
         &self,
         schema: &SchemaName,
         database_name: &str,
         container: &str,
     ) -> Result<Vec<String>, Error> {
         self.list_stored(schema, database_name, container)
-            .await
             .map_err(|err| Error::Backup(Box::new(err)))
     }
 
-    async fn load(
+    fn load(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -281,16 +266,14 @@ where
         name: &str,
     ) -> Result<Vec<u8>, Error> {
         self.load(schema, database_name, container, name)
-            .await
             .map_err(|err| Error::Backup(Box::new(err)))
     }
 }
 
-#[async_trait]
-impl<'a> BackupLocation for &'a Path {
+impl BackupLocation for Path {
     type Error = std::io::Error;
 
-    async fn store(
+    fn store(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -299,38 +282,33 @@ impl<'a> BackupLocation for &'a Path {
         object: &[u8],
     ) -> Result<(), Self::Error> {
         let container_folder = container_folder(self, schema, database_name, container);
-        tokio::fs::create_dir_all(&container_folder).await?;
-        tokio::fs::write(container_folder.join(name), object).await?;
+        std::fs::create_dir_all(&container_folder)?;
+        std::fs::write(container_folder.join(name), object)?;
 
         Ok(())
     }
 
-    async fn list_schemas(&self) -> Result<Vec<SchemaName>, Self::Error> {
-        iterate_directory(self, |entry, file_name| async move {
-            if entry.file_type().await?.is_dir() {
+    fn list_schemas(&self) -> Result<Vec<SchemaName>, Self::Error> {
+        iterate_directory(self, |entry, file_name| {
+            if entry.file_type()?.is_dir() {
                 if let Ok(schema_name) = SchemaName::parse_encoded(file_name.as_str()) {
                     return Ok(Some(schema_name));
                 }
             }
             Ok(None)
         })
-        .await
     }
 
-    async fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Self::Error> {
-        iterate_directory(
-            &schema_folder(self, schema),
-            |entry, file_name| async move {
-                if entry.file_type().await?.is_dir() && file_name != "_kv" {
-                    return Ok(Some(file_name));
-                }
-                Ok(None)
-            },
-        )
-        .await
+    fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Self::Error> {
+        iterate_directory(&schema_folder(self, schema), |entry, file_name| {
+            if entry.file_type()?.is_dir() && file_name != "_kv" {
+                return Ok(Some(file_name));
+            }
+            Ok(None)
+        })
     }
 
-    async fn list_stored(
+    fn list_stored(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -338,32 +316,30 @@ impl<'a> BackupLocation for &'a Path {
     ) -> Result<Vec<String>, Self::Error> {
         iterate_directory(
             &container_folder(self, schema, database_name, container),
-            |entry, file_name| async move {
-                if entry.file_type().await?.is_file() {
+            |entry, file_name| {
+                if entry.file_type()?.is_file() {
                     return Ok(Some(file_name));
                 }
                 Ok(None)
             },
         )
-        .await
     }
 
-    async fn load(
+    fn load(
         &self,
         schema: &SchemaName,
         database_name: &str,
         container: &str,
         name: &str,
     ) -> Result<Vec<u8>, Self::Error> {
-        tokio::fs::read(container_folder(self, schema, database_name, container).join(name)).await
+        std::fs::read(container_folder(self, schema, database_name, container).join(name))
     }
 }
 
-#[async_trait]
 impl BackupLocation for PathBuf {
     type Error = std::io::Error;
 
-    async fn store(
+    fn store(
         &self,
         schema: &SchemaName,
         database_name: &str,
@@ -371,56 +347,63 @@ impl BackupLocation for PathBuf {
         name: &str,
         object: &[u8],
     ) -> Result<(), Self::Error> {
-        BackupLocation::store(&*self, schema, database_name, container, name, object).await
+        BackupLocation::store(
+            self.as_path(),
+            schema,
+            database_name,
+            container,
+            name,
+            object,
+        )
     }
 
-    async fn list_schemas(&self) -> Result<Vec<SchemaName>, Self::Error> {
-        BackupLocation::list_schemas(&*self).await
+    fn list_schemas(&self) -> Result<Vec<SchemaName>, Self::Error> {
+        BackupLocation::list_schemas(self.as_path())
     }
 
-    async fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Self::Error> {
-        BackupLocation::list_databases(&*self, schema).await
+    fn list_databases(&self, schema: &SchemaName) -> Result<Vec<String>, Self::Error> {
+        BackupLocation::list_databases(self.as_path(), schema)
     }
 
-    async fn list_stored(
+    fn list_stored(
         &self,
         schema: &SchemaName,
         database_name: &str,
         container: &str,
     ) -> Result<Vec<String>, Self::Error> {
-        BackupLocation::list_stored(&*self, schema, database_name, container).await
+        BackupLocation::list_stored(self.as_path(), schema, database_name, container)
     }
 
-    async fn load(
+    fn load(
         &self,
         schema: &SchemaName,
         database_name: &str,
         container: &str,
         name: &str,
     ) -> Result<Vec<u8>, Self::Error> {
-        BackupLocation::load(&*self, schema, database_name, container, name).await
+        BackupLocation::load(self.as_path(), schema, database_name, container, name)
     }
 }
 
-async fn iterate_directory<
-    T,
-    F: FnMut(DirEntry, String) -> Fut,
-    Fut: Future<Output = Result<Option<T>, std::io::Error>>,
->(
+fn iterate_directory<T, F: FnMut(DirEntry, String) -> Result<Option<T>, std::io::Error>>(
     path: &Path,
     mut callback: F,
 ) -> Result<Vec<T>, std::io::Error> {
     let mut collected = Vec::new();
-    let mut directories =
-        if let Some(directories) = tokio::fs::read_dir(path).await.ignore_not_found()? {
-            directories
-        } else {
-            return Ok(collected);
-        };
+    let mut directories = if let Some(directories) = std::fs::read_dir(path).ignore_not_found()? {
+        directories
+    } else {
+        return Ok(collected);
+    };
 
-    while let Some(entry) = directories.next_entry().await.ignore_not_found()?.flatten() {
+    while let Some(entry) = directories
+        .next()
+        .map(IoResultExt::ignore_not_found)
+        .transpose()?
+        .flatten()
+    {
         if let Ok(file_name) = entry.file_name().into_string() {
-            if let Some(result) = callback(entry, file_name).await? {
+            if let Some(result) = callback(entry, file_name)? {
                 collected.push(result);
             }
         }
@@ -468,17 +451,19 @@ fn container_folder(
 #[cfg(test)]
 mod tests {
     use bonsaidb_core::{
-        connection::Connection as _,
+        connection::{Connection as _, StorageConnection as _},
         keyvalue::KeyValue,
         schema::SerializedCollection,
         test_util::{Basic, TestDirectory},
     };
 
-    use super::*;
-    use crate::config::{Builder, KeyValuePersistence, PersistenceThreshold, StorageConfiguration};
+    use crate::{
+        config::{Builder, KeyValuePersistence, PersistenceThreshold, StorageConfiguration},
+        Storage,
+    };
 
-    #[tokio::test]
-    async fn backup_restore() -> anyhow::Result<()> {
+    #[test]
+    fn backup_restore() -> anyhow::Result<()> {
         let backup_destination = TestDirectory::new("backup-restore.bonsaidb.backup");
 
         // First, create a database that we'll be restoring. `TestDirectory`
@@ -492,20 +477,16 @@ mod tests {
                         PersistenceThreshold::after_changes(2),
                     ]))
                     .with_schema::<Basic>()?,
-            )
-            .await?;
-            storage.create_database::<Basic>("basic", false).await?;
-            let db = storage.database::<Basic>("basic").await?;
-            let test_doc = db
-                .collection::<Basic>()
-                .push(&Basic::new("somevalue"))
-                .await?;
-            db.set_numeric_key("key1", 1_u64).await?;
-            db.set_numeric_key("key2", 2_u64).await?;
-            // This key will not be persisted right away.
-            db.set_numeric_key("key3", 3_u64).await?;
+            )?;
 
-            storage.backup(&*backup_destination.0).await.unwrap();
+            let db = storage.create_database::<Basic>("basic", false)?;
+            let test_doc = db.collection::<Basic>().push(&Basic::new("somevalue"))?;
+            db.set_numeric_key("key1", 1_u64).execute()?;
+            db.set_numeric_key("key2", 2_u64).execute()?;
+            // This key will not be persisted right away.
+            db.set_numeric_key("key3", 3_u64).execute()?;
+
+            storage.backup(&backup_destination.0).unwrap();
 
             test_doc
         };
@@ -513,27 +494,18 @@ mod tests {
         // `backup_destination` now contains an export of the database, time to try loading it:
         let database_directory = TestDirectory::new("backup-restore.bonsaidb");
         let restored_storage =
-            Storage::open(StorageConfiguration::new(&database_directory).with_schema::<Basic>()?)
-                .await?;
-        restored_storage
-            .restore(&*backup_destination.0)
-            .await
-            .unwrap();
+            Storage::open(StorageConfiguration::new(&database_directory).with_schema::<Basic>()?)?;
+        restored_storage.restore(&backup_destination.0).unwrap();
 
-        let db = restored_storage.database::<Basic>("basic").await?;
-        let doc = Basic::get(test_doc.id, &db)
-            .await?
-            .expect("Backed up document.not found");
+        let db = restored_storage.database::<Basic>("basic")?;
+        let doc = Basic::get(test_doc.id, &db)?.expect("Backed up document.not found");
         assert_eq!(doc.contents.value, "somevalue");
-        assert_eq!(db.get_key("key1").into_u64().await?, Some(1));
-        assert_eq!(db.get_key("key2").into_u64().await?, Some(2));
-        assert_eq!(db.get_key("key3").into_u64().await?, Some(3));
+        assert_eq!(db.get_key("key1").into_u64()?, Some(1));
+        assert_eq!(db.get_key("key2").into_u64()?, Some(2));
+        assert_eq!(db.get_key("key3").into_u64()?, Some(3));
 
         // Calling restore again should generate an error.
-        assert!(restored_storage
-            .restore(&*backup_destination.0)
-            .await
-            .is_err());
+        assert!(restored_storage.restore(&backup_destination.0).is_err());
 
         Ok(())
     }
