@@ -4,18 +4,16 @@ use std::{
     sync::Arc,
 };
 
-use async_lock::RwLock;
 use bonsaidb_core::{
     connection::Connection,
     keyvalue::Timestamp,
     schema::{view, CollectionName, ViewName},
 };
-use bonsaidb_utils::{fast_async_read, fast_async_write};
+use parking_lot::RwLock;
 
 use crate::{
     database::{keyvalue::ExpirationLoader, Database},
-    jobs::{manager::Manager, task::Handle},
-    tasks::compactor::Compactor,
+    tasks::{compactor::Compactor, handle::Handle, manager::Manager},
     views::{
         integrity_scanner::{IntegrityScan, IntegrityScanner, OptionalViewMapHandle},
         mapper::{Map, Mapper},
@@ -23,6 +21,14 @@ use crate::{
     Error,
 };
 
+/// Types related to defining [`Job`]s.
+pub mod handle;
+/// Types related to the job [`Manager`](manager::Manager).
+pub mod manager;
+mod traits;
+
+pub use self::traits::{Job, Keyed};
+
 mod compactor;
 mod task;
 
@@ -51,23 +57,23 @@ impl TaskManager {
         }
     }
 
-    pub async fn update_view_if_needed(
+    pub fn update_view_if_needed(
         &self,
         view: &dyn view::Serialized,
         database: &Database,
     ) -> Result<(), crate::Error> {
         let view_name = view.view_name();
-        if let Some(job) = self.spawn_integrity_check(view, database).await? {
-            job.receive().await??;
+        if let Some(job) = self.spawn_integrity_check(view, database) {
+            job.receive()??;
         }
 
         // If there is no transaction id, there is no data, so the view is "up-to-date"
-        if let Some(current_transaction_id) = database.last_transaction_id().await? {
+        if let Some(current_transaction_id) = database.last_transaction_id()? {
             let needs_reindex = {
                 // When views finish updating, they store the last transaction_id
                 // they mapped. If that value is current, we don't need to go
                 // through the jobs system at all.
-                let statuses = fast_async_read!(self.statuses);
+                let statuses = self.statuses.read();
                 if let Some(last_transaction_indexed) = statuses.view_update_last_status.get(&(
                     database.data.name.clone(),
                     view.collection(),
@@ -82,18 +88,15 @@ impl TaskManager {
             if needs_reindex {
                 let wait_for_transaction = current_transaction_id;
                 loop {
-                    let job = self
-                        .jobs
-                        .lookup_or_enqueue(Mapper {
-                            database: database.clone(),
-                            map: Map {
-                                database: database.data.name.clone(),
-                                collection: view.collection(),
-                                view_name: view_name.clone(),
-                            },
-                        })
-                        .await;
-                    let id = job.receive().await??;
+                    let job = self.jobs.lookup_or_enqueue(Mapper {
+                        database: database.clone(),
+                        map: Map {
+                            database: database.data.name.clone(),
+                            collection: view.collection(),
+                            view_name: view_name.clone(),
+                        },
+                    });
+                    let id = job.receive()??;
                     if wait_for_transaction <= id {
                         break;
                     }
@@ -104,131 +107,124 @@ impl TaskManager {
         Ok(())
     }
 
-    pub async fn key_value_expiration_loaded(&self, database: &Arc<Cow<'static, str>>) -> bool {
-        let statuses = fast_async_read!(self.statuses);
+    pub fn key_value_expiration_loaded(&self, database: &Arc<Cow<'static, str>>) -> bool {
+        let statuses = self.statuses.read();
         statuses.key_value_expiration_loads.contains(database)
     }
 
-    pub async fn view_integrity_checked(
+    pub fn view_integrity_checked(
         &self,
         database: Arc<Cow<'static, str>>,
         collection: CollectionName,
         view_name: ViewName,
     ) -> bool {
-        let statuses = fast_async_read!(self.statuses);
+        let statuses = self.statuses.read();
         statuses
             .completed_integrity_checks
-            .contains(&(database, collection.clone(), view_name))
+            .contains(&(database, collection, view_name))
     }
 
-    pub async fn spawn_integrity_check(
+    pub fn spawn_integrity_check(
         &self,
         view: &dyn view::Serialized,
         database: &Database,
-    ) -> Result<Option<Handle<OptionalViewMapHandle, Error, Task>>, crate::Error> {
+    ) -> Option<Handle<OptionalViewMapHandle, Error>> {
         let view_name = view.view_name();
-        if !self
-            .view_integrity_checked(
-                database.data.name.clone(),
-                view.collection(),
-                view_name.clone(),
-            )
-            .await
-        {
-            let job = self
-                .jobs
-                .lookup_or_enqueue(IntegrityScanner {
-                    database: database.clone(),
-                    scan: IntegrityScan {
-                        database: database.data.name.clone(),
-                        view_version: view.version(),
-                        collection: view.collection(),
-                        view_name,
-                    },
-                })
-                .await;
-            return Ok(Some(job));
+        if self.view_integrity_checked(
+            database.data.name.clone(),
+            view.collection(),
+            view_name.clone(),
+        ) {
+            None
+        } else {
+            let job = self.jobs.lookup_or_enqueue(IntegrityScanner {
+                database: database.clone(),
+                scan: IntegrityScan {
+                    database: database.data.name.clone(),
+                    view_version: view.version(),
+                    collection: view.collection(),
+                    view_name,
+                },
+            });
+            Some(job)
         }
-
-        Ok(None)
     }
 
-    pub async fn mark_integrity_check_complete(
+    pub fn mark_integrity_check_complete(
         &self,
         database: Arc<Cow<'static, str>>,
         collection: CollectionName,
         view_name: ViewName,
     ) {
-        let mut statuses = fast_async_write!(self.statuses);
+        let mut statuses = self.statuses.write();
         statuses
             .completed_integrity_checks
             .insert((database, collection, view_name));
     }
 
-    pub async fn mark_key_value_expiration_loaded(&self, database: Arc<Cow<'static, str>>) {
-        let mut statuses = fast_async_write!(self.statuses);
+    pub fn mark_key_value_expiration_loaded(&self, database: Arc<Cow<'static, str>>) {
+        let mut statuses = self.statuses.write();
         statuses.key_value_expiration_loads.insert(database);
     }
 
-    pub async fn mark_view_updated(
+    pub fn mark_view_updated(
         &self,
         database: Arc<Cow<'static, str>>,
         collection: CollectionName,
         view_name: ViewName,
         transaction_id: u64,
     ) {
-        let mut statuses = fast_async_write!(self.statuses);
+        let mut statuses = self.statuses.write();
         statuses
             .view_update_last_status
             .insert((database, collection, view_name), transaction_id);
     }
 
-    pub async fn spawn_key_value_expiration_loader(
+    pub fn spawn_key_value_expiration_loader(
         &self,
-        database: &crate::Database,
-    ) -> Option<Handle<(), Error, Task>> {
-        if self.key_value_expiration_loaded(&database.data.name).await {
+        database: &Database,
+    ) -> Option<Handle<(), Error>> {
+        if self.key_value_expiration_loaded(&database.data.name) {
             None
         } else {
-            Some(
-                self.jobs
-                    .lookup_or_enqueue(ExpirationLoader {
-                        database: database.clone(),
-                        launched_at: Timestamp::now(),
-                    })
-                    .await,
-            )
+            Some(self.jobs.lookup_or_enqueue(ExpirationLoader {
+                database: database.clone(),
+                launched_at: Timestamp::now(),
+            }))
         }
     }
 
-    pub async fn compact_collection(
+    pub fn spawn_compact_target(
         &self,
-        database: crate::Database,
+        database: Database,
+        target: compactor::Target,
+    ) -> Handle<(), Error> {
+        self.jobs
+            .lookup_or_enqueue(Compactor::target(database, target))
+    }
+
+    pub fn compact_collection(
+        &self,
+        database: Database,
         collection_name: CollectionName,
     ) -> Result<(), Error> {
         Ok(self
             .jobs
             .lookup_or_enqueue(Compactor::collection(database, collection_name))
-            .await
-            .receive()
-            .await??)
+            .receive()??)
     }
 
-    pub async fn compact_key_value_store(&self, database: crate::Database) -> Result<(), Error> {
+    pub fn compact_key_value_store(&self, database: Database) -> Result<(), Error> {
         Ok(self
             .jobs
             .lookup_or_enqueue(Compactor::keyvalue(database))
-            .await
-            .receive()
-            .await??)
+            .receive()??)
     }
 
-    pub async fn compact_database(&self, database: crate::Database) -> Result<(), Error> {
+    pub fn compact_database(&self, database: Database) -> Result<(), Error> {
         Ok(self
             .jobs
             .lookup_or_enqueue(Compactor::database(database))
-            .await
-            .receive()
-            .await??)
+            .receive()??)
     }
 }
